/*
 * Copyright (C) 2007 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <gmodule.h>
#include <glib.h>
#include <ohm-plugin.h>
#include <time.h>
#include <sys/stat.h>

#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <string.h>
#include <stdlib.h>

#define DCON_FREEZE_PATH "/sys/class/backlight/dcon-bl/device/freeze"
#define DCON_MODE_PATH   "/sys/class/backlight/dcon-bl/device/output"

enum {
	CONF_AC_STATE_CHANGED,
	CONF_BACKLIGHT_STATE_CHANGED,
	CONF_LID_STATE_CHANGED,
	CONF_TABLET_STATE_CHANGED,
	CONF_BRIGHTNESS_AC_CHANGED,
	CONF_BRIGHTNESS_BATTERY_CHANGED,
	CONF_BRIGHTNESS_IDLE_CHANGED,
	CONF_TIMEOUTS_POWERSAVE_CHANGED,
	CONF_TIMEOUTS_MOMENTARY_CHANGED,
	CONF_DCON_FREEZE_CHANGED,
	CONF_DCON_MODE_CHANGED,
	CONF_LAST
};

static gboolean
set_dcon_freeze(OhmPlugin *plugin, gint freeze)
{
	FILE *fp;

	g_debug("in %s, freeze is %d", G_STRFUNC, freeze);

	/* We start out set to 2, since we need Sugar's setting to 0
	 * or 1 to trigger a value change.  We don't want to write 2
	   to the file, though. */
	if (freeze == 2) {
		g_warning ("Leaving %s, freeze is unset", G_STRFUNC);
		return FALSE;
	}

	fp = g_fopen(DCON_FREEZE_PATH, "w");
	if (!fp) {
		g_warning ("Couldn't open %s", DCON_FREEZE_PATH);
		return FALSE;
	}

	if (g_fprintf(fp, freeze ? "1" : "0") < 0) {
		g_warning ("Couldn't write to %s", DCON_FREEZE_PATH);
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}

static gboolean
set_dcon_mode(OhmPlugin *plugin, gint mode)
{
	FILE *fp;

	g_debug("in %s, mode is %d", G_STRFUNC, mode);

	/* We start out set to 2, since we need Sugar's setting to 0
	 * or 1 to trigger a value change.  We don't want to write 2
	   to the file, though. */
	if (mode == 2) {
		g_warning ("Leaving %s, mode is unset", G_STRFUNC);
		return TRUE;
	}

	fp = g_fopen(DCON_MODE_PATH, "w");
	if (!fp) {
		g_warning ("Couldn't open %s", DCON_MODE_PATH);
		return FALSE;
	}

	if (g_fprintf(fp, mode ? "1" : "0") < 0) {
		g_warning ("Couldn't write to %s", DCON_MODE_PATH);
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}

static gint
backlight_get_brightness (OhmPlugin *plugin)
{
	gint level = -1;
	FILE *fp;
	char levelchar[4];
	
	g_debug("%s", G_STRFUNC);
	fp = g_fopen("/sys/class/backlight/dcon-bl/brightness", "r");
	if (!fp) {
		g_warning ("Couldn't open /sys/class/backlight/dcon-bl/brightness.");
		return -1;
	}

	while (fgets(levelchar, sizeof(levelchar), fp) != NULL) {
		g_debug ("Level is %s.", levelchar);
		level = atoi(levelchar);
		g_debug ("Level is %d.", level);
		break;
	}

	fclose(fp);
	
	return level;
}

static void 
reset_idle(OhmPlugin *plugin)
{
	gint new_timeout;
	g_debug("in %s", G_STRFUNC);
	
	ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
	ohm_plugin_conf_set_key (plugin, "idle.reset", 1);
	ohm_plugin_conf_set_key (plugin, "idle.state", 0);

	ohm_plugin_conf_get_key (plugin, "timeouts.timer_momentary", &new_timeout);
	ohm_plugin_conf_set_key (plugin, "idle.timeout", new_timeout);
	ohm_plugin_conf_set_key (plugin, "timeouts.momentary", 0);
	ohm_plugin_conf_set_key (plugin, "timeouts.powerdown", 0);
}

static void
reset_brightness (OhmPlugin *plugin)
{
	gint onac;
	gint value_current;
	gint value_ac;
	gint value_battery;
	gint value_idle;
	gint value_new = -1;
	gint value_user;
	
	g_debug("%s", G_STRFUNC);

	ohm_plugin_conf_get_key (plugin, "acadapter.state", &onac);
	ohm_plugin_conf_get_key (plugin, "display.value_ac", &value_ac);
	ohm_plugin_conf_get_key (plugin, "display.value_battery", &value_battery);
	ohm_plugin_conf_get_key (plugin, "display.value_idle", &value_idle);
	ohm_plugin_conf_get_key (plugin, "display.value_user", &value_user);

	value_current = backlight_get_brightness(plugin);

	if (value_current == -1) {
		g_warning("Can't read brightness!");
		return;
	}

	if (value_current != value_user) {
		g_debug ("Setting value_new to user %d", value_user);
		value_new = value_user;
	}
	else {
		if (onac == TRUE) {
			value_new = value_ac;
		} else {
			value_new = value_battery;
		}
	}
	
	g_debug("%s: acadapter.state = %d, setting brightness %d", G_STRFUNC, onac, value_new);

	/* dim screen to idle brightness */
	ohm_plugin_conf_set_key (plugin, "backlight.direct_set_brightness", value_new);
}

static gboolean
cpu_is_busy (OhmPlugin *plugin)
{
#define IDLE_PERCENT 10

	gchar *output = NULL;
	GError *error = NULL;
	gint idle;

	g_debug("in %s", G_STRFUNC);

	if (!g_spawn_command_line_sync ("/bin/sh -c \"vmstat 1 2 | tail -n1 | awk '{print $15}'\"", &output, NULL, NULL, &error)) {
		g_warning ("Error spawning: %s\n", error->message);
		g_error_free (error);
	}

	idle = atoi(output);
	if (idle < IDLE_PERCENT) {
		g_debug("idle %d is less than IDLE_PERCENT; inhibiting", idle);
		return TRUE;
	} else {
		g_debug("idle %d is more than IDLE_PERCENT; proceeding", idle);
	}

	return FALSE;
}

static gboolean
check_model(OhmPlugin *plugin)
{
	char model[5];
	FILE *fp;
	
	fp = g_fopen("/ofw/model", "r");
	if (!fp) {
		g_warning ("Couldn't open /ofw/model.");
		return FALSE;
	}

	while (fgets(model, sizeof(model), fp) != NULL) {
		g_debug ("Model is %s.", model);
	}

	if (g_ascii_strncasecmp(model, "C", 1) == 0) {
		g_debug ("C-series, returning TRUE.");
		return TRUE;
	}
	else {
		g_debug ("Not C-series, returning FALSE.");
		return FALSE;
	}
}


static gboolean 
okay_to_suspend (gpointer data)
{
#define TTY1 "/dev/tty1"
#define TTY2 "/dev/tty2"
#define SUSPEND_TAG_FILE "/etc/ohm/inhibit-suspend"
#define SUSPEND_IDLE_TAG_FILE "/etc/ohm/inhibit-idle-suspend"

	gint activity_inhibit;
	gint lidshut;
	gint value_idle;
	gint value_user;
	gint value_current;
	gint state;
	gint idle;
	gboolean supports_suspend;
	struct stat tty1;
	struct stat tty2;
	struct stat stat_buf;
	time_t curtime;

        OhmPlugin *plugin = (OhmPlugin *) data;
	
	g_debug("%s", G_STRFUNC);

	ohm_plugin_conf_get_key (plugin, "idle.state", &idle);
	if (idle != 1) {
		/* Something's wrong; we must have reset out of momentary. */
		ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
		return FALSE;
	}

	/* check for idleness inhibit file */
	if (stat(SUSPEND_IDLE_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not idle suspending; idle inhibit file found.");
		ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
		return FALSE;
	}
	
	/* Are we inhibited by a sugar activity? */
	ohm_plugin_conf_get_key (plugin, "suspend.inhibit", &activity_inhibit);
	if (activity_inhibit == 1) {
		g_debug("%s: inhibiting due to activity dbus message", G_STRFUNC);
		return FALSE;
	}
	
	/* We don't want an old inhibit to carry over. */
	ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 0);
	
	ohm_plugin_conf_get_key (plugin, "backlight.state", &state);
	if (state == 0) {
                g_debug("%s: state is 0", G_STRFUNC);
		/* work round a idletime bugs */
		return FALSE;
	}

	/* if lid shut */
	ohm_plugin_conf_get_key (plugin, "button.lid", &lidshut);
	if (lidshut == 1) {
	        g_debug("%s: lid shut", G_STRFUNC);
		/* we've already turned off the screen */
		return FALSE;
	}

	supports_suspend = check_model(plugin);
	if (!supports_suspend) {
		ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
 		return FALSE;
	}

	/* check for general inhibit file */
	if (stat(SUSPEND_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not suspending; inhibit file found.");
		ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
		return FALSE;
	}
	
	curtime = time(NULL);
	g_debug ("time now is %lu", (unsigned long) curtime);

	if (stat(TTY1, &tty1) >= 0) {
		g_debug ("tty1 time is %lu", (unsigned long) tty1.st_mtime);
		if (curtime-tty1.st_mtime >= 0 && curtime-tty1.st_mtime < 60) {
			/* We check for >= 0 so that a future filesystem doesn't inhibit. */
			g_debug("tty1 is in use; inhibiting screen dim.");
			ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
			return FALSE;
		}
	}

	if (stat(TTY2, &tty2) >= 0) {
		g_debug ("tty2 time is %lu", (unsigned long) tty2.st_mtime);
		if (curtime-tty2.st_mtime >= 0 && curtime-tty2.st_mtime < 60) {
			g_debug("tty2 is in use; inhibiting screen dim.");
			ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
			return FALSE;
		}
	}

	if (cpu_is_busy(plugin)) {
		g_debug("cpu busy; inhibiting screen dim.");
		ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 1);
		return FALSE;
	}

	/* dim screen to idle brightness */
	/* First, see whether the current brightness is user-set. */ 
	value_current = backlight_get_brightness(plugin);
	
	if (value_current == -1) {
		g_warning ("Can't read brightness!");
		return FALSE;
	}
	
	ohm_plugin_conf_get_key (plugin, "display.value_idle", &value_idle);
	ohm_plugin_conf_get_key (plugin, "display.value_user", &value_user);

	if (value_current != value_user) {
		g_debug ("Brightness (%d) isn't user set (%d)", value_current, value_user);
	}

	if (value_user > value_idle) {
		g_debug ("User brightness %d is > idle brightness %d", value_user, value_idle);
		g_debug ("%s: Setting brightness to %d", G_STRFUNC, value_idle);
		ohm_plugin_conf_set_key (plugin, "backlight.direct_set_brightness", value_idle);
	}
	else {
		g_debug ("User brightness %d is <= idle brightness %d", value_user, value_idle);
		g_debug ("Not dimming because value_user %d is already less than value_idle %d.", value_user, value_idle);
	}

	/* needed to stop the callback from firing again */
	return FALSE;
}

/* todo, we need to inhibit the screen from turning off */
/* todo, we need to inhibit the screen from dimming */
static void
brightness_momentary (OhmPlugin *plugin, gboolean is_idle)
{
	g_debug("in %s", G_STRFUNC);
	/* if not idle any more */
	if (is_idle == FALSE) {
		g_debug("%s: is_idle=FALSE< restting brighness", G_STRFUNC);
		// cjb
		reset_brightness (plugin);
		reset_idle (plugin);
		return;
	}

	g_timeout_add (3000, okay_to_suspend, plugin);

}

static void
backlight_powersave (OhmPlugin *plugin, gboolean is_idle)
{
	gint lidshut;
	gint state;

	ohm_plugin_conf_get_key (plugin, "backlight.state", &state);
	if (is_idle && state == 0) {
	        g_debug("%s: is_idle 0", G_STRFUNC);
		/* work round a idletime bugs */
		return;
	}

	/* if lid shut */
	ohm_plugin_conf_get_key (plugin, "button.lid", &lidshut);
	if (lidshut == 1) {
	        g_debug("%s: lidshut 1", G_STRFUNC);
		/* we've already turned off the screen */
		return;
	}

	/* if not idle any more */
	if (is_idle == FALSE) {
	        g_debug("%s: if not idle any more", G_STRFUNC);
		ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
		reset_brightness (plugin);
		return;
	}

}

static void
lid_closed (OhmPlugin *plugin, gboolean is_closed)
{
	/* just turn on DCON */
	if (is_closed == FALSE) {
		ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
		return;
	}
		
	/* just turn off dcon unconditionally */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 0);
}

/**
 * plugin_notify:
 * @plugin: This class instance
 *
 * Notify the plugin that a key marked with ohm_plugin_conf_interested ()
 * has it's value changed.
 * An enumerated numeric id rather than the key is returned for processing speed.
 */
static void
plugin_notify (OhmPlugin *plugin, gint id, gint value)
{
	switch (id) {
	case CONF_BRIGHTNESS_AC_CHANGED:
	case CONF_BRIGHTNESS_BATTERY_CHANGED:
	case CONF_BRIGHTNESS_IDLE_CHANGED:
		reset_brightness (plugin);
		break;
	case CONF_TIMEOUTS_POWERSAVE_CHANGED:
		backlight_powersave (plugin, (value == 1));
		break;
	case CONF_LID_STATE_CHANGED:
		lid_closed (plugin, (value == 1));
		break;
	case CONF_TABLET_STATE_CHANGED:
		break;
	case CONF_BACKLIGHT_STATE_CHANGED:
		reset_brightness (plugin);
		break;
	case CONF_TIMEOUTS_MOMENTARY_CHANGED:
		brightness_momentary (plugin, (value == 1));
		break;
	case CONF_DCON_FREEZE_CHANGED:
		set_dcon_freeze (plugin, value);
		break;
	case CONF_DCON_MODE_CHANGED:
		set_dcon_mode (plugin, value);
		break;
	}
}

OHM_PLUGIN_DESCRIPTION (
	"Display",			/* description */
	"0.0.1",			/* version */
	"richard@hughsie.com",		/* author */
	OHM_LICENSE_LGPL,		/* license */
	NULL,				/* initialize */
	NULL,				/* destroy */
	plugin_notify			/* notify */
);

OHM_PLUGIN_SUGGESTS (
	"timeouts",
	"acadapter",
	"buttons",
	"xrandr",
	"backlight"
);

OHM_PLUGIN_INTERESTED (
	{"acadapter.state", CONF_AC_STATE_CHANGED},
	{"backlight.state", CONF_BACKLIGHT_STATE_CHANGED},
	{"button.lid", CONF_LID_STATE_CHANGED},
	{"button.tablet", CONF_TABLET_STATE_CHANGED},
	{"timeouts.powersave", CONF_TIMEOUTS_POWERSAVE_CHANGED},
	{"timeouts.momentary", CONF_TIMEOUTS_MOMENTARY_CHANGED},
	{"display.value_ac", CONF_BRIGHTNESS_AC_CHANGED},
	{"display.value_battery", CONF_BRIGHTNESS_BATTERY_CHANGED},
	{"display.value_idle", CONF_BRIGHTNESS_IDLE_CHANGED},
	{"display.dcon_freeze", CONF_DCON_FREEZE_CHANGED},
	{"display.dcon_mode", CONF_DCON_MODE_CHANGED}
);
