/*
 * Copyright (C) 2007 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <sys/stat.h>
#include <time.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <string.h>

#include <gmodule.h>
#include <glib.h>
#include <ohm-plugin.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus.h>

#define HAL_DBUS_SERVICE		"org.freedesktop.Hal"
#define HAL_DBUS_INTERFACE_POWER	"org.freedesktop.Hal.Device.SystemPowerManagement"
#define HAL_ROOT_COMPUTER		"/org/freedesktop/Hal/devices/computer"
#define NM_DBUS_SERVICE 		"org.freedesktop.NetworkManager"
#define NM_DBUS_INTERFACE		"org.freedesktop.NetworkManager"
#define NM_DBUS_PATH			"/org/freedesktop/NetworkManager"

#define DCON_SLEEP_PATH 		"/sys/class/backlight/dcon-bl/device/sleep"
#define WLAN_ENABLED_PATH		"/sys/power/wlan-enabled"

#define SUSPEND_TAG_FILE		"/etc/ohm/inhibit-suspend"
#define SUSPEND_IDLE_TAG_FILE 		"/etc/ohm/inhibit-idle-suspend"

enum {
	CONF_BUTTON_POWER_CHANGED,
	CONF_BUTTON_LID_CHANGED,
	CONF_TIMEOUTS_POWERDOWN_CHANGED,
	CONF_INHIBIT_CHANGED,
	CONF_AUTOMATIC_PM,
	CONF_EXTREME_PM,
	CONF_LAST
};


static gboolean
set_wakeup(OhmPlugin *plugin, gchar device[], gboolean value)
{
	FILE *fp;
	gchar filename[60] = "/sys/power/wakeup_events/";

	strncat(filename, device, 30);
	g_debug("in %s, filename is %s", G_STRFUNC, filename);

	fp = g_fopen(filename, "w");
	if (!fp) {
		g_warning ("Couldn't open %s", filename);
		return FALSE;
	}

	if (g_fprintf(fp, value? "1" : "0") < 0) {
		g_warning ("Couldn't write to %s", filename);
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}


static gboolean
set_dcon_sleep(OhmPlugin *plugin, gboolean sleep)
{
	FILE *fp;
	
	g_debug("in %s", G_STRFUNC);

	/* Turn off the backlight when sleeping.  We can't do this with
	   DPMS, because it would wake us back up on a keystroke. */ 
        
	fp = g_fopen(DCON_SLEEP_PATH, "w");
	if (!fp) {
		g_warning ("Couldn't open %s", DCON_SLEEP_PATH);
		return FALSE;
	}

	if (g_fprintf(fp, sleep ? "1" : "0") < 0) {
		g_warning ("Couldn't write to %s", DCON_SLEEP_PATH);
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}

static gboolean
set_wlan_power(OhmPlugin *plugin, gboolean enabled)
{
	FILE *fp;
	g_debug("in %s", G_STRFUNC);

	/* Turn off power to the libertas 8388 via the EC. */
	fp = g_fopen(WLAN_ENABLED_PATH, "w");
	if (!fp) {
		g_warning ("Couldn't open %s", WLAN_ENABLED_PATH);
		return FALSE;
	}

	if (g_fprintf(fp, enabled ? "1" : "0") < 0) {
		g_warning ("Couldn't write to %s", WLAN_ENABLED_PATH);
		return FALSE;
	}

	fclose(fp);
	return TRUE;
}

static void
networkmanager_up(gboolean wake)
{
	DBusGConnection *conn;
	DBusGProxy *proxy;

	conn = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);

	proxy = dbus_g_proxy_new_for_name (conn,
					   NM_DBUS_SERVICE,
					   NM_DBUS_PATH,
					   NM_DBUS_INTERFACE);

	if (wake == TRUE) {
		g_debug ("nm wake");
		dbus_g_proxy_call_no_reply (proxy, "wake",
					    G_TYPE_INVALID);
	}
	else {
		g_debug ("nm sleep");
		dbus_g_proxy_call_no_reply (proxy, "sleep",
					    G_TYPE_INVALID);	
	}
	
	g_object_unref (proxy);
}

static gboolean
check_model(OhmPlugin *plugin)
{
	char model[5];
	FILE *fp;
	
	fp = g_fopen("/ofw/model", "r");
	if (!fp) {
		g_warning ("Couldn't open /ofw/model.");
		return FALSE;
	}

	while (fgets(model, sizeof(model), fp) != NULL) {
		g_debug ("Model is %s.", model);
	}

        fclose(fp);

	if (g_ascii_strncasecmp(model, "C", 1) == 0) {
		g_debug ("C-series, returning TRUE.");
		return TRUE;
	}
	else {
		g_debug ("Not C-series, returning FALSE.");
		return FALSE;
	}
}

static gboolean
do_suspend(OhmPlugin *plugin)
{
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	fp = g_fopen("/sys/power/state", "w");
	if (!fp) {
		g_warning ("Couldn't open /sys/power/state.");
		return FALSE;
	}

	if (g_fprintf(fp, "mem") < 0) {
		g_warning ("Couldn't write to /sys/power/state.");
		return FALSE;
	}

	fclose(fp);

	g_debug("should have suspended");
	return TRUE;
}

static gboolean
pre_sleep(OhmPlugin *plugin)
{
	g_debug("in %s", G_STRFUNC);

	/* Turn off the backlight when sleeping.  We can't do this with
	   DPMS, because it would wake us back up on a keystroke. */ 

	set_dcon_sleep(plugin, TRUE);

	set_wakeup(plugin, "all", FALSE);
	
	return TRUE;
}

static gboolean
pre_suspend(OhmPlugin *plugin)
{
  	g_debug("in %s", G_STRFUNC);

	/* For an idle suspend, turn off unwanted events. */
	set_wakeup(plugin, "battery_soc", FALSE);
	set_wakeup(plugin, "ac_power", FALSE);

        return TRUE;
}

static void 
reset_idle(OhmPlugin *plugin)
{
	gint new_timeout;
	g_debug("in %s", G_STRFUNC);
	
	/* We've resumed now.  Bring the backlight back immediately. */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
	ohm_plugin_conf_set_key (plugin, "idle.reset", 1);
	ohm_plugin_conf_set_key (plugin, "idle.state", 0);

	ohm_plugin_conf_get_key (plugin, "timeouts.timer_momentary", &new_timeout);
	ohm_plugin_conf_set_key (plugin, "idle.timeout", new_timeout);
	ohm_plugin_conf_set_key (plugin, "timeouts.momentary", 0);
	ohm_plugin_conf_set_key (plugin, "timeouts.powerdown", 0);
}

static gboolean
post_suspend(OhmPlugin *plugin)
{
	g_debug("in %s", G_STRFUNC);

	/* For an idle suspend, re-enable battery 1% events and AC.
	 * Without this, the battery status applet would not update.
	 */
	set_wakeup(plugin, "battery_soc", TRUE);
	set_wakeup(plugin, "ac_power", TRUE);

	reset_idle(plugin);
	return TRUE;
}

static gboolean
post_sleep(OhmPlugin *plugin)
{
	char source[256];
	FILE *fp;

	g_debug("in %s", G_STRFUNC);
	
	fp = g_fopen("/sys/power/wakeup-source", "r");
	if (!fp) {
		g_debug ("Couldn't open /sys/power/wakeup-source.");
		return TRUE;
	}

	while (fgets(source, sizeof(source), fp) != NULL) {
		g_debug ("Source was %s.", source);
	}

	fclose(fp);

	/* Wake up fully. */
	set_dcon_sleep(plugin, FALSE);

	set_wakeup(plugin, "all", TRUE);

	reset_idle(plugin);
	return TRUE;
}

static gboolean
disable_wireless_unless_mesh (OhmPlugin *plugin)
{
	gchar *output = NULL;
	GError *error = NULL;

	/* Turn off the wireless, unless we're using the mesh.
	   I wish there was a dbus interface to this. */
	if (!g_spawn_command_line_sync ("/bin/sh -c \"nm-tool | grep msh0 -A4 | grep Active | awk '{print $2}'\"", &output, NULL, NULL, &error)) {
		g_warning ("Error spawning: %s\n", error->message);
		g_error_free (error);
	}

	g_debug("Output is: %s", output);

	if (g_ascii_strncasecmp(output, "yes", 3) == 0) {
		/* Mesh is in use.  We can't disable wireless. */
		g_debug("Not disabling wifi, mesh is in use.");
	}
	else if (g_ascii_strncasecmp(output, "no", 2) == 0) {
		g_debug("Disabling wifi, mesh is not in use.");

		/* We return TRUE if we disabled wireless, else FALSE. */
		set_wlan_power(plugin, FALSE);
		return TRUE;
	}
	else {
		g_debug("Strange output from nm-tool: %s", output);
	}

	return FALSE;
}

static void
power_button_pressed (OhmPlugin *plugin)
{
	gboolean supports_suspend;
	gboolean disabled_wireless = FALSE;

	struct stat stat_buf;

	supports_suspend = check_model(plugin);
	if (!supports_suspend)
		return;

	/* check for inhibit file */
	if (stat(SUSPEND_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not suspending; inhibit file found.");
		return;
	}

	disabled_wireless = disable_wireless_unless_mesh(plugin);
	pre_sleep(plugin);
	do_suspend(plugin);
	post_sleep(plugin);
	if (disabled_wireless) {
		/* If we turned off wireless going down, turn it back on. */
		set_wlan_power(plugin, TRUE);
	}
	else {
		/* If wifi has been up, kick the NetworkManager connection
		   since we've probably been asleep for a while. */
		networkmanager_up(FALSE);
		networkmanager_up(TRUE);
	}
}

static void
lid_closed (OhmPlugin *plugin, gboolean is_closed)
{
	if (is_closed == FALSE)
		return;

	/* Call power_button_pressed to avoid duplicating code. */
	power_button_pressed(plugin);
}

static gboolean
tty_in_use (OhmPlugin *plugin)
{
#define TTY1 "/dev/tty1"
#define TTY2 "/dev/tty2"

	struct stat tty1;
	struct stat tty2;
	time_t curtime;

	g_debug("in %s", G_STRFUNC);

	curtime = time(NULL);
	g_debug ("time now is %lu", (unsigned long) curtime);
	
	if (stat(TTY1, &tty1) >= 0) {
		g_debug ("tty1 time is %lu", (unsigned long) tty1.st_mtime);
		if (curtime-tty1.st_mtime < 60) {
			g_debug("tty1 is in use; inhibiting suspend.");
			return TRUE;
		}
	}

	if (stat(TTY2, &tty2) >= 0) {
		g_debug ("tty2 time is %lu", (unsigned long) tty2.st_mtime);
		if (curtime-tty2.st_mtime < 60) {
			g_debug("tty2 is in use; inhibiting suspend.");
			return TRUE;
		}
	}

	return FALSE;
}	
	
static gboolean 
system_is_idle (OhmPlugin *plugin)
{
	gint ohm_inhibit;
	gint activity_inhibit;
	struct stat stat_buf;

	gboolean supports_suspend;
	supports_suspend = check_model(plugin);
	if (!supports_suspend)
		return FALSE;

	g_debug("in %s", G_STRFUNC);

	/* check for inhibit file */
	if (stat(SUSPEND_TAG_FILE, &stat_buf) >= 0) {
		/* it exists */
		g_debug("not suspending; inhibit file found.");
		return FALSE;
	}

        /* check for idleness inhibit file */
        if (stat(SUSPEND_IDLE_TAG_FILE, &stat_buf) >= 0) {
                /* it exists */
                g_debug("not idle suspending; idle inhibit file found.");
                return FALSE;
        }

	/* check system inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &ohm_inhibit);

	/* check activity inhibit - this is broken as any client can unref all */
	ohm_plugin_conf_get_key (plugin, "suspend.inhibit", &activity_inhibit);

	if (ohm_inhibit == 1) {
		g_debug ("inhibiting suspend: ohm inhibit set");
		return FALSE;
	}

	if (activity_inhibit == 1) {
		g_debug ("inhibiting suspend: activity inhibit set");
		return FALSE;
	}

	if (tty_in_use(plugin))
		return FALSE;

	pre_suspend(plugin);
	do_suspend(plugin);
	post_suspend(plugin);
	return TRUE;
}

/**
 * plugin_notify:
 * @plugin: This class instance
 *
 * Notify the plugin that a key marked with ohm_plugin_conf_interested ()
 * has it's value changed.
 * An enumerated numeric id rather than the key is returned for processing speed.
 */
static void
plugin_notify (OhmPlugin *plugin, gint id, gint value)
{
	gint inhibit_suspend = 0;
	struct stat stat_buf;
	char wlanstate[5];
	FILE *wlanfp;
	FILE *idlefp;

	if (id == CONF_BUTTON_LID_CHANGED) {
		if (value == 1) {
			lid_closed (plugin, TRUE);
		} else {
			lid_closed (plugin, FALSE);
		}
	} else if (id == CONF_BUTTON_POWER_CHANGED) {
		/* only match on button press, not release */
		if (value == 1) {
			power_button_pressed (plugin);
		}
	} else if (id == CONF_AUTOMATIC_PM) {
		g_debug ("conf_automatic_pm changed");
		if (value == 1) {
			g_debug ("value is 1");
			if (stat(SUSPEND_IDLE_TAG_FILE, &stat_buf) >= 0) {
				/* it exists */
				g_debug("removing idle inhibit file");
				if (g_unlink(SUSPEND_IDLE_TAG_FILE) != 0)
					g_warning("Couldn't unlink %s!", SUSPEND_IDLE_TAG_FILE);
			}
		}
		else if (value == 0) {
			g_debug ("value is 0");
			if (stat(SUSPEND_IDLE_TAG_FILE, &stat_buf) < 0) {
				/* it doesn't exist */
				g_debug("creating idle inhibit file");
				idlefp = g_fopen(SUSPEND_IDLE_TAG_FILE, "w");
				if (!idlefp)
					g_warning ("Couldn't touch %s!", SUSPEND_IDLE_TAG_FILE);
			}
		}
	} else if (id == CONF_EXTREME_PM) {
		g_debug ("conf_extreme_pm changed");
		wlanfp = g_fopen("/sys/power/wlan-enabled", "r");
		if (!wlanfp) {
			g_debug ("Couldn't open /sys/power/wlan-enabled.");
			return;
		}
		while (fgets(wlanstate, sizeof(wlanstate), wlanfp) != NULL) {
			g_debug ("wlanstate is %s.", wlanstate);
		}
		fclose(wlanfp);

		if (value == 1) {
			g_debug("value is 1");

			if (g_ascii_strncasecmp(wlanstate, "1", 1) == 0) {
			    /* We're being told to disable the enabled radio. */
			    g_debug("Disabling the enabled radio.");
			    set_wlan_power(plugin, 0);
			    return;
			}
			if (g_ascii_strncasecmp(wlanstate, "0", 1) == 0) {
			    /* We're being told to disable the disabled radio. */
			    g_debug("Not disabling the disabled radio.");
			    return;
			}
		}
		else if (value == 0) {
			g_debug("value is 0");
			if (g_ascii_strncasecmp(wlanstate, "1", 1) == 0) {
			    /* We're being told to enable the enabled radio. */
			    g_debug("Not enabling the enabled radio.");
			    return;
			}
			if (g_ascii_strncasecmp(wlanstate, "0", 1) == 0) {
			    /* We're being told to disable the disabled radio. */
			    g_debug("Enabling the disabled radio.");
			    set_wlan_power(plugin, 1);
			    return;
			}
		}
	} else if (id == CONF_TIMEOUTS_POWERDOWN_CHANGED) {
		/* only match on idle, not reset */
		g_debug ("powerdown changed");
		if (value == 1) {
			g_debug ("powerdown is 1, suspending");
			ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &inhibit_suspend);
			if (inhibit_suspend == 1) {
				g_debug ("inhibit_suspend is 1, resetting");
				ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 0);
				reset_idle(plugin);	
				return;
			}
			ohm_plugin_conf_get_key (plugin, "suspend.fixme_inhibit", &inhibit_suspend);
			if (inhibit_suspend == 0) {
				g_debug ("inhibit_suspend was 0, going ahead");
				if (!system_is_idle (plugin)) {
					g_debug ("didn't suspend -- resetting idle");
					reset_idle(plugin);
					return;
				}
			}
		}
		else if (value == 0) {
			g_debug ("powerdown is 0");
		}
	}
}

static void
plugin_init(OhmPlugin *plugin)
{
	/* Set by OHM itself. */
	ohm_plugin_conf_set_key (plugin, "suspend.fixme_inhibit", 0);
	/* Set externally by activity developers. */
	ohm_plugin_conf_set_key (plugin, "suspend.inhibit", 0);
}

OHM_PLUGIN_DESCRIPTION (
	"suspend",			/* description */
	"0.0.1",			/* version */
	"richard@hughsie.com",		/* author */
	OHM_LICENSE_LGPL,		/* license */
	plugin_init,				/* initialize */
	NULL,				/* destroy */
	plugin_notify			/* notify */
);

OHM_PLUGIN_SUGGESTS (
	"timeouts",
	"buttons"
);

OHM_PLUGIN_INTERESTED (
	{"button.power", CONF_BUTTON_POWER_CHANGED},
	{"button.lid", CONF_BUTTON_LID_CHANGED},
	{"timeouts.powerdown", CONF_TIMEOUTS_POWERDOWN_CHANGED},
	{"suspend.fixme_inhibit", CONF_INHIBIT_CHANGED},
	{"suspend.automatic_pm", CONF_AUTOMATIC_PM},
	{"suspend.extreme_pm", CONF_EXTREME_PM}
);
